#include "stdlib.h"
#include "stdio.h"
#include "stdbool.h"
#include "assert.h"
#include "unistd.h"
#include "time.h"
#include <SQLite3/sqlite3.h>

// Get the maximum number of workers in parallel
int GetNbMaxWorker(sqlite3* const db);
// Get the number of running task
int GetNbRunningTask(sqlite3* const db);
// Process a task
bool ProcessTask(sqlite3* const db);
// Get the reference of a pending task
int GetRefPendingTask(sqlite3* const db);
// Try to assign a worker to a task
bool AssignWorkerToTask(
  sqlite3* const db, int const refTask, char const* const key);
// Get the input
int GetInput(sqlite3* const db, int const refTask);
// Run the task
int RunTask(sqlite3* const db, int const refTask, int const input);
// Save the result of the task
void SaveResult(sqlite3* const db, int const refTask, int const output);
// Sleep for the given delay in millisecond
int SleepMs(int32_t const delayMs);

int main(int argc, char** argv) {
  // If the number of arguments is not the expected one
  if(argc != 2) {
    // Print an error message and terminate with the failure code
    fprintf(stderr, "usage: worker <dbObfuscation>\n");
    return EXIT_FAILURE;
  }
  // Reference to the db obfuscation parameter
  char const* const dbObfuscation = argv[1];
  // Create the path to the database
  char path[100];
  snprintf(path, 100, "./TaskQueue.%s.db", dbObfuscation);
  sqlite3* db = NULL;
  int ret = sqlite3_open(path, &db);
  assert(ret == SQLITE_OK);
  // Get the max number of workers
  int const nbMaxWorker = GetNbMaxWorker(db);
  printf("nbMaxWorker %d\n", nbMaxWorker);
  // Main loop 
  bool flagContinue = true;
  while(flagContinue) {
    // Get the number of running tasks
    int nbRunningTask = GetNbRunningTask(db);
    printf("nbRunningTask %d\n", nbRunningTask);
    // If the max number of workers is reached
    if(nbRunningTask >= nbMaxWorker) {
      // Terminate the worker
      flagContinue = false;
    } else {
      // Try to process a pending task
      flagContinue = ProcessTask(db);
    }
  }
  // Free memory
  ret = sqlite3_close(db);
  assert(ret == SQLITE_OK);
  // Normal termination, return the success code
  return EXIT_SUCCESS;
}

// Get the maximum number of workers in parallel
int GetNbMaxWorker(sqlite3* const db) {
  sqlite3_stmt *stmt = NULL;
  int ret = 0;
  do {
    ret = sqlite3_prepare_v2(db,
      "SELECT NbMaxWorker FROM TaskConfig WHERE Reference=1;", -1, &stmt, NULL);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  assert(ret == SQLITE_OK);
  do {
    ret = sqlite3_step(stmt);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  assert(ret == SQLITE_ROW);
  int const nbMaxWorker = sqlite3_column_int(stmt, 0);
  sqlite3_finalize(stmt);
  return nbMaxWorker;
}

// Get the number of running task
int GetNbRunningTask(sqlite3* const db) {
  sqlite3_stmt *stmt = NULL;
  int ret = 0;
  do {
    ret = sqlite3_prepare_v2(db,
      "SELECT COUNT(*) FROM Task WHERE Status=1;", -1, &stmt, NULL);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  assert(ret == SQLITE_OK);
  do {
    ret = sqlite3_step(stmt);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  assert(ret == SQLITE_ROW);
  int const nbRunningTask = sqlite3_column_int(stmt, 0);
  sqlite3_finalize(stmt);
  return nbRunningTask;
}

// Process a task
bool ProcessTask(sqlite3* const db) {
  // Get the reference of a pending task
  int refTask = GetRefPendingTask(db);
  if(refTask != -1) {
    char key[30];
    snprintf(key, 30, "%d_%d", getpid(), refTask);
    // Try to assign the worker to the task
    bool const taskActuallyTaken = AssignWorkerToTask(db, refTask, key);
    if(taskActuallyTaken) {
      // Get the input
      int input = GetInput(db, refTask);
      // Run the task
      int const output = RunTask(db, refTask, input);
      // Save the result
      SaveResult(db, refTask, output);
    }
    // Continue processing task
    return true;
  } else {
    // No pending task, stop
    return false;
  }
}

// Get the reference of a pending task
int GetRefPendingTask(sqlite3* const db) {
  sqlite3_stmt *stmt = NULL;
  int ret = 0;
  do {
    ret = sqlite3_prepare_v2(db,
      "SELECT Reference FROM Task WHERE Status=0 LIMIT 1", -1, &stmt, NULL);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  assert(ret == SQLITE_OK);
  do { 
    ret = sqlite3_step(stmt);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  int refTask = -1;
  if(ret == SQLITE_ROW) refTask = sqlite3_column_int(stmt, 0);
  sqlite3_finalize(stmt);
  return refTask;
}

// Try to assign a worker to a task
bool AssignWorkerToTask(
  sqlite3* const db, int const refTask, char const* const key) {
  // Try to set the key and status of the task
  sqlite3_stmt *stmt = NULL;
  int ret = 0;
  do {
    ret = sqlite3_prepare_v2(db,
      "UPDATE Task SET Status=1, Key=?1, WorkerMsg='0%' "
      "WHERE Status=0 AND Reference=?2",
      -1, &stmt, NULL);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  assert(ret == SQLITE_OK);
  sqlite3_bind_text(stmt, 1, key, -1, SQLITE_STATIC);
  sqlite3_bind_int(stmt, 2, refTask);
  do {
    ret = sqlite3_step(stmt);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  assert(ret == SQLITE_DONE);
  sqlite3_finalize(stmt);
  // Check that the task has been taken
  do {
    ret = sqlite3_prepare_v2(db,
      "SELECT COUNT(*) FROM Task WHERE Status=1 AND Key=?1 AND Reference=?2",
      -1, &stmt, NULL);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  assert(ret == SQLITE_OK);
  sqlite3_bind_text(stmt, 1, key, -1, SQLITE_STATIC);
  sqlite3_bind_int(stmt, 2, refTask);
  do {
    ret = sqlite3_step(stmt);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  assert(ret == SQLITE_ROW);
  int const nb = sqlite3_column_int(stmt, 0);
  sqlite3_finalize(stmt);
  return (nb == 1);
}

// Get the input
int GetInput(sqlite3* const db, int const refTask) {
  sqlite3_stmt *stmt = NULL;
  int ret = 0;
  do {
    ret = sqlite3_prepare_v2(db,
      "SELECT Input FROM Task WHERE Reference=?1;", -1, &stmt, NULL);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  assert(ret == SQLITE_OK);
  sqlite3_bind_int(stmt, 1, refTask);
  do {
    ret = sqlite3_step(stmt);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  assert(ret == SQLITE_ROW);
  int const input = sqlite3_column_int(stmt, 0);
  sqlite3_finalize(stmt);
  return input;
}

// Run the task
int RunTask(sqlite3* const db, int const refTask, int const input) {
  // Dummy operation as a proof of work and sleep to create artificial load
  int const output = input + 1;
  sleep(1);
  sqlite3_stmt *stmt = NULL;
  int ret = 0;
  do {
    ret = sqlite3_prepare_v2(db,
      "UPDATE Task SET WorkerMsg='100%' WHERE Reference=?1",
      -1, &stmt, NULL);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  assert(ret == SQLITE_OK);
  sqlite3_bind_int(stmt, 1, refTask);
  do {
    ret = sqlite3_step(stmt);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  assert(ret == SQLITE_DONE);
  sqlite3_finalize(stmt);
  return output;
}

// Save the result of the task
void SaveResult(sqlite3* const db, int const refTask, int const output) {
  sqlite3_stmt *stmt = NULL;
  int ret = 0;
  do {
    ret = sqlite3_prepare_v2(db,
      "UPDATE Task SET Status=2, Output=?1 WHERE Reference=?2",
      -1, &stmt, NULL);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  assert(ret == SQLITE_OK);
  sqlite3_bind_int(stmt, 1, output);
  sqlite3_bind_int(stmt, 2, refTask);
  do {
    ret = sqlite3_step(stmt);
    if(ret == SQLITE_BUSY) SleepMs(10);
  } while(ret == SQLITE_BUSY);
  assert(ret == SQLITE_DONE);
  sqlite3_finalize(stmt);
}

int SleepMs(int32_t const delayMs) {
  struct timespec ts = {
    .tv_sec = delayMs / 1000,
    .tv_nsec = (delayMs % 1000) * 1000000,
  };
  return nanosleep(&ts, &ts);
}
